"""
bibliograpy command entrypoint
"""
import logging

from argparse import ArgumentParser, Namespace

from bibliograpy.process import _process

LOG = logging.getLogger(__name__)


def _info(ns: Namespace):
    """info
    """
    LOG.info("info %s", ns)

DEFAULT_FILE = "bibliograpy.yaml"
DEFAULT_ENCODING = 'utf-8'
DEFAULT_OUTPUT_DIR = '.'
DEFAULT_OUTPUT_FILE = 'bibliography.py'


def _create_parser() -> ArgumentParser:

    # parse argument line
    parser = ArgumentParser(description='Bibliography management.')

    subparsers = parser.add_subparsers(dest='CMD', help='available commands')

    subparsers.add_parser('info', help='get general info')

    process = subparsers.add_parser('process', help='generates bibliograpy source bibliography')
    process.add_argument('file',
                         nargs='?',
                         help="path to the bibliograpy configuration file",
                         default=DEFAULT_FILE)
    process.add_argument('--encoding', '-e',
                         nargs='?',
                         help='the bibliograpy configuration file encoding (default to utf-8)',
                         default=DEFAULT_ENCODING)
    process.add_argument('--output-dir', '-O',
                         nargs='?',
                         help='the source bibliograpy file output directory',
                         default=DEFAULT_OUTPUT_DIR)
    process.add_argument('--output-file', '-o',
                         nargs='?',
                         help='the source bibliograpy output file name',
                         default=DEFAULT_OUTPUT_FILE)
    process.add_argument('--scope', '-s',
                         nargs='?',
                         help='the scope name, must be consistent with --scope-import')
    process.add_argument('--init-scope', '-S',
                         nargs='?',
                         help='the scope import line')

    return parser


def entrypoint():
    """The pyenvs command entrypoint."""

    commands = {
        'info': _info,
        'process': _process
    }

    ns: Namespace = _create_parser().parse_args()

    commands[ns.CMD](ns)



"""
bibliograpy process module
"""
import json
import logging
from argparse import Namespace
from pathlib import Path
from typing import Any

import bibtexparser
import yaml
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser.bwriter import BibTexWriter

from bibliograpy.api import TYPES
from bibliograpy.api import Reference

LOG = logging.getLogger(__name__)


def _process(ns: Namespace):
    """config
    """
    LOG.info("dependencies")

    in_extension = ns.file.split('.')[-1]
    output_dir = Path(Path.cwd(), ns.output_dir)
    output_file = ns.output_file
    out_extension = output_file.split('.')[-1]
    scope_symbol = ns.scope if 'scope' in ns else None
    init_scope = ns.init_scope if 'init_scope' in ns else None

    LOG.info('open configuration file %s', ns.file)
    with open(ns.file, encoding=ns.encoding) as s:
        content = _read(s, extension=in_extension)

        with open(Path(output_dir, output_file), 'w', encoding=ns.encoding) as o:
            _write(o, extension=out_extension, content=content, scope_symbol=scope_symbol, init_scope=init_scope)

def _read(s, extension: str) -> list:
    """Reads the input bibliography file content."""

    if extension == 'yml':
        return yaml.safe_load(s)

    if extension == 'json':
        return json.load(s)

    if extension == 'bib':
        meta = {}
        content = []
        for e in bibtexparser.load(s).entries:
            meta['entry_type'] = e['ENTRYTYPE']
            meta[Reference.CITE_KEY_FIELD] = e['ID']
            del e['ENTRYTYPE']
            del e['ID']
            content.append({**meta, **e})
        return content

    raise ValueError(f'unsupported configuration format {extension}')

def _write(o, extension: str, content: list, scope_symbol: str | None, init_scope: str | None):
    """Writes the bibliography in the format specified by the provided extension."""

    if extension == 'py':

        scope: dict[str, Any] = {}

        o.write('from bibliograpy.api import *\n')
        o.write('\n')

        if init_scope is not None:
            o.write(f'{init_scope}\n')
            o.write('\n')

        for ref in content:
            if ref['entry_type'] in TYPES:
                o.write(f"{TYPES[ref['entry_type']].from_dict(ref, scope).to_py(scope_symbol=scope_symbol)}\n")
    elif extension in ['yml', 'yaml']:
        yaml.dump(content, o, sort_keys=False)
    elif extension in ['bib']:
        scope: dict[str, Any] = {}
        entries = []
        for ref in content:
            if ref['entry_type'] in TYPES:
                entries.append(TYPES[ref['entry_type']].from_dict(ref, scope).to_bib())
        db = BibDatabase()
        db.entries = entries
        writer = BibTexWriter()
        writer.order_entries_by = None
        bibtexparser.dump(bib_database=db, bibtex_file=o, writer=writer)
    elif extension in ['json']:
        json.dump(content, fp=o, sort_keys=False)



"""Test module for bibliograpy tool"""

from argparse import Namespace
from pathlib import Path
import pydoc
import sys
import time

import yaml

import pytest

from bibliograpy.api import cite

from bibliograpy.process import _process

def _input_file(file: str) -> str:
    """Les fichiers d'entrée se trouvent à côté des scripts de test."""
    return str(Path(Path(__file__).parent, file))

def _output_file(file: str) -> str:
    """Les fichiers de sortie sont générés relativement à l'endroit où la commande est exécutée."""
    return str(Path(Path.cwd(), file))


def test_process_yml_to_yml():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.yml'),
                       output_file=_input_file('test_process_yml_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_yml_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [{
            'entry_type': 'misc',
            'cite_key': 'nasa',
            'title': 'NASA'
        },{
            'entry_type': 'misc',
            'cite_key': 'iau',
            'title': 'International Astronomical Union'
        }]

def test_process_yml_to_bib():
    """test process from a yml bibliography to a bibtex bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.yml'),
                       output_file=_input_file('test_process_yml_to_bib.bib'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_yml_to_bib.bib'), encoding='utf-8') as s:
        content = s.read()
        assert content == """@misc{nasa,
 title = {NASA}
}

@misc{iau,
 title = {International Astronomical Union}
}
"""

def test_process_yml_to_json():
    """test process from a yml bibliography to a bibtex bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.yml'),
                       output_file=_input_file('test_process_yml_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_yml_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"entry_type": "misc", "cite_key": "nasa", "title": "NASA"}, '
                           '{"entry_type": "misc", "cite_key": "iau", "title": "International Astronomical Union"}]')

def test_process_json_to_yml():
    """test process from a json bibliography to a yml bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.json'),
                       output_file=_input_file('test_process_json_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_json_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [{
            'entry_type': 'misc',
            'cite_key': 'nasa',
            'title': 'NASA'
        },{
            'entry_type': 'misc',
            'cite_key': 'iau',
            'title': 'International Astronomical Union'
        }]

def test_process_json_to_bib():
    """test process from a json bibliography to a bibtex bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.json'),
                       output_file=_input_file('test_process_json_to_bib.bib'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_json_to_bib.bib'), encoding='utf-8') as s:
        content = s.read()
        assert content == """@misc{nasa,
 title = {NASA}
}

@misc{iau,
 title = {International Astronomical Union}
}
"""

def test_process_json_to_json():
    """test process from a json bibliography to a json bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.json'),
                       output_file=_input_file('test_process_json_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_json_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"entry_type": "misc", "cite_key": "nasa", "title": "NASA"}, '
                           '{"entry_type": "misc", "cite_key": "iau", "title": "International Astronomical Union"}]')

def test_process_bib_to_yml():
    """test process from a bib bibliography to a yml bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.json'),
                       output_file=_input_file('test_process_bib_to_yml.yml'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_bib_to_yml.yml'), encoding='utf-8') as s:
        content = yaml.safe_load(s)
        assert content == [{
            'entry_type': 'misc',
            'cite_key': 'nasa',
            'title': 'NASA'
        },{
            'entry_type': 'misc',
            'cite_key': 'iau',
            'title': 'International Astronomical Union'
        }]

def test_process_bib_to_bib():
    """test process from a bibtex bibliography to a bibtex bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.bib'),
                       output_file=_input_file('test_process_bib_to_bib.bib'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_bib_to_bib.bib'), encoding='utf-8') as s:
        content = s.read()
        assert content == """@misc{nasa,
 title = {NASA}
}

@misc{iau,
 title = {International Astronomical Union}
}
"""

def test_process_bib_to_json():
    """test process from a bibtex bibliography to a json bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.bib'),
                       output_file=_input_file('test_process_bib_to_json.json'),
                       encoding='utf-8',
                       output_dir='.'))

    with open(_input_file('test_process_bib_to_json.json'), encoding='utf-8') as s:
        content = s.read()
        assert content == ('[{"entry_type": "misc", "cite_key": "nasa", "title": "NASA"}, '
                           '{"entry_type": "misc", "cite_key": "iau", "title": "International Astronomical Union"}]')

def test_process_yml_to_yml_astroloj():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('astroloj.json'),
                       output_file=_input_file('astroloj.py'),
                       encoding='utf-8',
                       output_dir='.'))

def test_process_yml_to_yml_cosmoloj():
    """test process from a yml bibliography to a yml bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('cosmoloj.json'),
                       output_file=_input_file('cosmoloj.py'),
                       encoding='utf-8',
                       output_dir='.'))

def test_process_yml_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.yml'),
                       output_file=_input_file('test_process_yml_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_process_yml_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_process_json_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.json'),
                       output_file=_input_file('test_process_json_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_process_json_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_process_bib_to_py():
    """test process from a yml bibliography to a py source bibliography"""

    _process(Namespace(CMD='process',
                       file=_input_file('mini.bib'),
                       output_file=_input_file('test_process_bib_to_py.py'),
                       encoding='utf-8',
                       output_dir='.'))

    time.sleep(1) # wait for the bibliography source file to be generated

    from test_process_bib_to_py import IAU, NASA

    @cite(IAU, NASA)
    def bib_ref_foo():
        """ma doc avec plusieurs références en varargs"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * International Astronomical Union [iau]
    * NASA [nasa]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_bibliograpy

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * International Astronomical Union [iau]
    * NASA [nasa]
""")

def test_process_input_file_not_found():
    """test process input file not found"""

    with pytest.raises(FileNotFoundError) as e:
        with open(_output_file('not_existing_file.yml'), encoding='utf-8') as s:
            yaml.safe_load(s)

    assert e.value.args[0] == 2
    assert e.value.args[1] == "No such file or directory"



"""Test module for bibliograpy"""
import sys
import pydoc

import pytest

from bibliograpy.api import cite, Misc, TechReport, Reference, ReferenceBuilder, inbook

SCOPE: dict[str, Reference] = {}


IAU = Misc.generic(cite_key='iau',
                   title='International Astronomical Union',
                   institution='IAU',
                   scope=SCOPE)


IAU_2006_B1 = TechReport.generic(
    cite_key='iau_2006_b1',
    author='',
    crossref='iau',
    title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
    year=2006,
    scope=SCOPE)


def test_to_source_bib():
    """test to python source bib serialization"""
    assert (IAU_2006_B1.to_py(scope_symbol=None) ==
"""
IAU_2006_B1 = TechReport.generic(cite_key='iau_2006_b1',
                                 author='',
                                 crossref=IAU,
                                 title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
                                 year=2006)""")


def test_to_source_bib_with_scope():
    """test to python source bib serialization"""
    assert (IAU_2006_B1.to_py(scope_symbol='SCOPE') ==
"""
IAU_2006_B1 = TechReport.generic(cite_key='iau_2006_b1',
                                 author='',
                                 crossref=IAU,
                                 title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
                                 year=2006,
                                 scope=SCOPE)""")


def test_builtin_reference_decorator():
    """test build-in reference decorator with a single reference, an array of references and references in varargs"""

    @cite(IAU_2006_B1)
    def bib_ref():
        """ma doc"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref) ==
"""Python Library Documentation: function bib_ref in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf()
    ma doc

    Bibliography: Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
""")
    else:
        assert (pydoc.render_doc(bib_ref) ==
"""Python Library Documentation: function bib_ref in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf()
    ma doc
    
    Bibliography: Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
""")

    @cite([IAU_2006_B1, IAU])
    def bib_ref_foo():
        """ma doc avec plusieurs références"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références

    Bibliography:

    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")
    else:
        assert (pydoc.render_doc(bib_ref_foo) ==
"""Python Library Documentation: function bib_ref_foo in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_f\bfo\boo\bo()
    ma doc avec plusieurs références
    
    Bibliography:
    
    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")

    @cite(IAU_2006_B1, IAU)
    def bib_ref_bar():
        """ma doc avec plusieurs références en varargs"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref_bar) ==
"""Python Library Documentation: function bib_ref_bar in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_b\bba\bar\br()
    ma doc avec plusieurs références en varargs

    Bibliography:

    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")
    else:
        assert (pydoc.render_doc(bib_ref_bar) ==
"""Python Library Documentation: function bib_ref_bar in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf_\b_b\bba\bar\br()
    ma doc avec plusieurs références en varargs
    
    Bibliography:
    
    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")


def test_custom_reference_builder():
    """test custom reference builder"""

    def _ref_formatter(r: Reference) -> str:
        return f"{r.title} [{r.cite_key}]"

    def custom_wrapper(refs: list[Reference]) -> str:
        if len(refs) == 1:
            return f"\n\nBibliographie : {_ref_formatter(refs[0])}\n"

        result = "\n\nBibliographie :\n\n"
        for r in refs:
            result += f"* {_ref_formatter(r)}\n"
        return result

    ref = ReferenceBuilder(reference_wrapper=custom_wrapper)

    @ref(IAU_2006_B1, IAU)
    def tatafr():
        """ma doc avec plusieurs références en varargs"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs

    Bibliographie :

    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")
    else:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs
    
    Bibliographie :
    
    * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
    * International Astronomical Union [iau]
""")


def test_parameterized_default_reference_builder():
    """test parameterized default reference builder"""

    def _formatter(ref: Reference) -> str:
        base = f'{ref.title} [{ref.cite_key}]'
        if ref.crossref:
            return base + f' -> [{ref.crossref}]'
        return base

    ref = ReferenceBuilder.default(prefix='Références bibliographiques :',
                                   itemize='++',
                                   reference_formatter=_formatter)

    @ref(IAU_2006_B1, IAU)
    def tatafr():
        """ma doc avec plusieurs références en varargs"""


    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs

    Références bibliographiques :

    ++ Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1] -> [iau]
    ++ International Astronomical Union [iau]
""")
    else:
        assert (pydoc.render_doc(tatafr) ==
"""Python Library Documentation: function tatafr in module test_api

t\bta\bat\bta\baf\bfr\br()
    ma doc avec plusieurs références en varargs
    
    Références bibliographiques :
    
    ++ Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1] -> [iau]
    ++ International Astronomical Union [iau]
""")


def test_mandatory_field():
    """test mandatory field management"""
    with pytest.raises(ValueError) as e:
        TechReport.generic(
            cite_key='iau_2006_b1',
            author='',
            title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
            year=2006)
    assert e.value.args[0] == 'missing mandatory field for TechReport iau_2006_b1'


def test_cross_reference():
    """test cross reference hierarchy management"""
    scope = {}
    assert len(scope) == 0

    iau_org = Misc.generic(cite_key='iau',
                           institution='Internation Astronomical Union',
                           author='iau',
                           crossref='no_ref',
                           scope=scope)
    assert len(scope) == 1
    assert 'iau' in scope
    assert scope['iau'] is iau_org
    assert iau_org.institution == 'Internation Astronomical Union'
    assert iau_org.author == 'iau'

    iau_author = Misc.generic(cite_key='iau_author', author='IAU', crossref='iau', scope=scope)
    assert len(scope) == 2
    assert 'iau_author' in scope
    assert scope['iau_author'] is iau_author
    assert iau_author.institution is None
    assert iau_author.author == 'IAU'
    assert iau_author.cross_resolved().institution == 'Internation Astronomical Union'

    iau_2006 = TechReport.generic(
        cite_key='iau_2006_b1',
        crossref='iau_author',
        title='Adoption of the P03 Precession Theory and Definition of the Ecliptic',
        year=2006,
        scope=scope)
    assert len(scope) == 3
    assert 'iau_2006_b1' in scope
    assert scope['iau_2006_b1'] is iau_2006
    assert iau_2006.institution is None
    assert iau_2006.cross_resolved().institution == 'Internation Astronomical Union'
    assert iau_2006.author is None
    assert iau_2006.cross_resolved().author == 'IAU'


def test_specific_entry_type_decorator():
    """test build-in reference decorator with a single reference, an array of references and references in varargs"""

    @inbook(crossref=IAU_2006_B1, title="mon inbook", pages=235, publisher='', author='auteur')
    def bib_ref():
        """ma doc"""

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(bib_ref) ==
"""Python Library Documentation: function bib_ref in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf()
    ma doc

    Bibliography: mon inbook
""")
    else:
        assert (pydoc.render_doc(bib_ref) ==
"""Python Library Documentation: function bib_ref in module test_api

b\bbi\bib\bb_\b_r\bre\bef\bf()
    ma doc
    
    Bibliography: mon inbook
""")


def test_builtin_reference_decorator_class_usage():
    """test build-in reference decorator with a single reference, an array of references and references in varargs"""

    @cite(IAU_2006_B1)
    class BibRef:
        """ma doc"""

        def bib(self):
            """ma doc bib"""
            return "foo"

        def ref(self):
            """ma doc ref"""
            return "bar"

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(BibRef) ==
                """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |
 |  Bibliography: Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |
 |  Methods defined here:
 |
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    elif sys.version_info.minor >= 11:
        assert (pydoc.render_doc(BibRef) ==
                """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |  
 |  Bibliography: Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    else:
        assert (pydoc.render_doc(BibRef) ==
                """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |  
 |  Bibliography: Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables (if defined)
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object (if defined)
""")

    @cite([IAU_2006_B1, IAU])
    class BibRefFoo:
        """ma doc"""

        def bib(self):
            """ma doc bib"""
            return "foo"

        def ref(self):
            """ma doc ref"""
            return "bar"

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(BibRefFoo) ==
                """Python Library Documentation: class BibRefFoo in module test_api

class B\bBi\bib\bbR\bRe\bef\bfF\bFo\boo\bo(builtins.object)
 |  ma doc
 |
 |  Bibliography:
 |
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |
 |  Methods defined here:
 |
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    elif sys.version_info.minor >= 11:
        assert (pydoc.render_doc(BibRefFoo) ==
                """Python Library Documentation: class BibRefFoo in module test_api

class B\bBi\bib\bbR\bRe\bef\bfF\bFo\boo\bo(builtins.object)
 |  ma doc
 |  
 |  Bibliography:
 |  
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    else:
        assert (pydoc.render_doc(BibRefFoo) ==
                """Python Library Documentation: class BibRefFoo in module test_api

class B\bBi\bib\bbR\bRe\bef\bfF\bFo\boo\bo(builtins.object)
 |  ma doc
 |  
 |  Bibliography:
 |  
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables (if defined)
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object (if defined)
""")

    @cite(IAU_2006_B1, IAU)
    class BibRefBar:
        """ma doc"""

        def bib(self):
            """ma doc bib"""
            return "foo"

        def ref(self):
            """ma doc ref"""
            return "bar"

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(BibRefBar) ==
                """Python Library Documentation: class BibRefBar in module test_api

class B\bBi\bib\bbR\bRe\bef\bfB\bBa\bar\br(builtins.object)
 |  ma doc
 |
 |  Bibliography:
 |
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |
 |  Methods defined here:
 |
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    elif sys.version_info.minor >= 11:
        assert (pydoc.render_doc(BibRefBar) ==
                """Python Library Documentation: class BibRefBar in module test_api

class B\bBi\bib\bbR\bRe\bef\bfB\bBa\bar\br(builtins.object)
 |  ma doc
 |  
 |  Bibliography:
 |  
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    else:
        assert (pydoc.render_doc(BibRefBar) ==
                """Python Library Documentation: class BibRefBar in module test_api

class B\bBi\bib\bbR\bRe\bef\bfB\bBa\bar\br(builtins.object)
 |  ma doc
 |  
 |  Bibliography:
 |  
 |  * Adoption of the P03 Precession Theory and Definition of the Ecliptic [iau_2006_b1]
 |  * International Astronomical Union [iau]
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables (if defined)
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object (if defined)
""")


def test_specific_entry_type_decorator_class_usage():
    """test build-in reference decorator with a single reference, an array of references and references in varargs"""

    @inbook(crossref=IAU_2006_B1, title="mon inbook", pages=235, publisher='', author='auteur')
    class BibRef:
        """ma doc"""

        def bib(self):
            """ma doc bib"""
            return "foo"

        def ref(self):
            """ma doc ref"""
            return "bar"

    if sys.version_info.minor >= 12:
        assert (pydoc.render_doc(BibRef) ==
                """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |
 |  Bibliography: mon inbook
 |
 |  Methods defined here:
 |
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    elif sys.version_info.minor >= 11:
        assert (pydoc.render_doc(BibRef) ==
                    """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |  
 |  Bibliography: mon inbook
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object
""")
    else:
        assert (pydoc.render_doc(BibRef) ==
                    """Python Library Documentation: class BibRef in module test_api

class B\bBi\bib\bbR\bRe\bef\bf(builtins.object)
 |  ma doc
 |  
 |  Bibliography: mon inbook
 |  
 |  Methods defined here:
 |  
 |  b\bbi\bib\bb(self)
 |      ma doc bib
 |  
 |  r\bre\bef\bf(self)
 |      ma doc ref
 |  
 |  ----------------------------------------------------------------------
 |  Data descriptors defined here:
 |  
 |  _\b__\b_d\bdi\bic\bct\bt_\b__\b_
 |      dictionary for instance variables (if defined)
 |  
 |  _\b__\b_w\bwe\bea\bak\bkr\bre\bef\bf_\b__\b_
 |      list of weak references to the object (if defined)
""")



"""Bibliograpy API module."""
from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import Callable, Any
import logging

LOG = logging.getLogger(__name__)

_ANONYM_CITE_KEY = ""

@dataclass(frozen=True)
class NonStandard:
    """Non-standard bibtex bibliography reference fields."""

    doi: str | None = None
    """DOI number"""

    issn: str | None = None
    """ISSN number"""

    eissn: str | None = None
    """ISSN number"""

    isbn: str | None = None
    """ISBN number"""

    url: str | None = None
    """URL of a web page"""

    @staticmethod
    def from_dict(source: dict) -> NonStandard | None:
        """Builds a non-standard reference field set from a dict."""
        if any(f in source for f in ['doi', 'issn', 'eisssn', 'isbn', 'url']):
            return NonStandard(
                doi=source['doi'] if 'doi' in source else None,
                issn=source['issn'] if 'issn' in source else None,
                eissn=source['eissn'] if 'eissn' in source else None,
                isbn=source['isbn'] if 'isbn' in source else None,
                url=source['url'] if 'url' in source else None)
        return None


    def to_py(self) -> str:
        """Serialization of the non-standard reference field set in processed python code."""

        base = f"{type(self).__name__}("

        fields = []
        for f in dataclasses.fields(type(self)):
            value = getattr(self, f.name)

            if value is not None:
                fields.append(f"{f.name}='{value}'")

        return f"{base}{', '.join(fields)})"


@dataclass(frozen=True, repr=False)
class Reference:
    """A bibliography reference."""

    CITE_KEY_FIELD = 'cite_key'
    NON_STANDARD_FIELD = 'non_standard'
    SCOPE_FIELD = 'scope'
    CROSSREF_FIELD = 'crossref'

    cite_key: str

    address: str | None
    """address of the publisher or the institution
    
    not used in article, misc and unpublished
    optional everywhere else
    https://www.bibtex.com/f/address-field/"""

    annote: str | None
    """an annotation
    
    https://www.bibtex.com/f/annote-field/"""

    author: str | None
    """ist of authors of the work
    
    optional for booklet, manual and misc
    required everywhere else
    https://www.bibtex.com/f/author-field/"""

    booktitle: str | None
    """title of the book
    
    required for incollection and inproceedings
    not used everywhere else
    https://www.bibtex.com/f/booktitle-field/"""

    chapter: str | None
    """number of a chapter in a book
    
    required for inbook and incollection
    not used everywhere else
    https://www.bibtex.com/f/chapter-field/"""

    crossref: str | None
    """The database key of the entry being cross referenced."""

    edition: str | None
    """edition number of a book
    
    optional for book, inbook, incollection and manual
    not used everywhere else
    https://www.bibtex.com/f/edition-field/"""

    editor: str | None
    """list of editors of a book
    
    required for book and inbook
    optional for incollection and inproceedings
    not used everywhere else
    https://www.bibtex.com/f/editor-field/"""

    howpublished: str | None
    """a publication notice for unusual publications
    
    optional for booklet and misc
    not used everywhere else
    https://www.bibtex.com/f/howpublished-field/"""

    institution: str | None
    """name of the institution that published and/or sponsored the report
    
    required for techreport
    not used everywhere else
    https://www.bibtex.com/f/institution-field/
    """

    journal: str | None
    """name of the journal or magazine the article was published in
    
    required for article
    not used everywhere else
    https://www.bibtex.com/f/journal-field/
    """

    month: str | None
    """the month during the work was published
    
    optional
    https://www.bibtex.com/f/month-field/"""

    note: str | None
    """
    notes about the reference
    
    required for unpublished
    optional everywhere else
    https://www.bibtex.com/f/note-field/"""

    number: str | int | None
    """number of the report or the issue number for a journal article
    
    optional for article, book, inbook, incollection, inproceedings and techreport
    not used everywhere else
    https://www.bibtex.com/f/number-field/"""

    organization: str | None
    """name of the institution that organized or sponsored the conference or that published the manual
    
    optional for inproceedings and manual
    not used everywhere else
    https://www.bibtex.com/f/organization-field/"""

    pages: str | int | None
    """page numbers or a page range
    
    required for inbook
    optional for article, incollection and inproceedings
    not used everywhere else
    https://www.bibtex.com/f/pages-field/"""

    publisher: str | None
    """name of the publisher
    
    required for book, inbook and incollection
    optional for inproceedings
    not used everywhere else
    https://www.bibtex.com/f/publisher-field/"""

    school: str | None
    """name of the university or degree awarding institution
    
    required for masterthesis and phdthesis
    not used everywhere else
    https://www.bibtex.com/f/school-field/"""

    series: str | None
    """name of the series or set of books
    
    optional for book, inbook, incollection and inproceedings
    not used everywhere else
    https://www.bibtex.com/f/series-field/"""

    title: str | None
    """title of the work
    
    optional for misc
    required everywhere else
    https://www.bibtex.com/f/title-field/"""

    type: str | None
    """type of the technical report or thesis
    
    optional for inbook, incollection, masterthesis and techreport
    not used everywhere else
    https://www.bibtex.com/f/type-field/"""

    volume: str | int | None
    """volume number
    
    optional for article, book, inbook, incollection and inproceedings
    not used everywhere else
    https://www.bibtex.com/f/volume-field/"""

    year: str | int | None
    """year the book was published
    
    required for article, book, inbook, incollection, inproceedings, masterthesis, phdthesis, techreport
    optional for booklet, misc and unpublished
    not used for manual
    https://www.bibtex.com/f/year-field/"""

    non_standard: NonStandard | None
    """Non standard fields."""

    scope: dict[str, Reference] | None
    """Environnement de résolution des références croisées."""

    def _hierarchy(self) -> list[Reference]:
        """Calcul de la hiérarchie par références croisées."""
        current: Reference | None = self

        hierarchy: list[Reference] = [current]
        while current is not None:
            if current.crossref is not None and current.scope is not None and current.crossref in current.scope:
                parent: Reference = current.scope[current.crossref]
                hierarchy.append(parent)
                current = parent
            else:
                current = None
        return hierarchy

    def cross_resolved(self) -> Reference:
        """Calcul de la référence héritant des champs des références croisées parentes."""
        hierarchy = self._hierarchy()

        if len(hierarchy) == 1:
            return self

        resolved_standard_dict: dict[str, Any] = {}
        for f in dataclasses.fields(type(self)):

            if f.name in [Reference.NON_STANDARD_FIELD, Reference.SCOPE_FIELD]:
                continue

            for i in hierarchy:
                v = getattr(i, f.name)
                if v is not None:
                    resolved_standard_dict[f.name] = v
                    break

        resolved_non_standard_dict: dict[str, Any] = {}
        for f in dataclasses.fields(NonStandard):

            for i in hierarchy:
                if i.non_standard is not None:
                    v = getattr(i.non_standard, f.name)
                    if v is not None:
                        resolved_non_standard_dict[f.name] = v
                        break

        resolved_standard_dict[Reference.NON_STANDARD_FIELD] = NonStandard.from_dict(resolved_non_standard_dict)

        # il ne faut pas ajouter dans le scope cette instance résolue par références croisées car sa clef étant la même
        # que celle de la référence explicite, les deux entreraient en conflit dans le scope
        # seule la référence explicite doit être ajoutée au scope
        # si on souhaite disposer des champs hérités par références croisées, il faut utiliser l'instance résolue
        return type(self).from_dict(source=resolved_standard_dict,
                                    scope=None)
    @staticmethod
    def to_source_symbol(cite_key: str) -> str:
        """Produces a Python symbol for the reference."""
        return cite_key.upper()

    def to_py(self, scope_symbol: str | None) -> str:
        """Serialization of the reference in processed python code."""

        base = f"{Reference.to_source_symbol(self.cite_key)} = {type(self).__name__}.generic("

        fields = []
        for f in dataclasses.fields(type(self)):

            if Reference.SCOPE_FIELD == f.name:
                continue

            value = getattr(self, f.name)

            if f.name == Reference.CROSSREF_FIELD and value is not None:
                fields.append(f"{f.name}={Reference.to_source_symbol(value)}")
            elif isinstance(value, str):
                if "'" in value:
                    fields.append(f'{f.name}="{value}"')
                else:
                    fields.append(f"{f.name}='{value}'")
            elif isinstance(value, NonStandard):
                fields.append(f'{f.name}={value.to_py()}')
            elif value is not None:
                fields.append(f'{f.name}={value}')

        if scope_symbol is not None:
            fields.append(f'{Reference.SCOPE_FIELD}={scope_symbol}')

        # argument indentation management
        sep = ',\n'
        for _ in range(len(base)):
            sep += ' '

        return f"\n{base}{sep.join(fields)})"

    def to_bib(self) -> dict:
        """converts to a bibtex parser dict"""
        result = {}
        for f in dataclasses.fields(type(self)):

            if Reference.SCOPE_FIELD == f.name:
                continue

            if Reference.CITE_KEY_FIELD == f.name:
                field_name = 'ID'
            else:
                field_name = f.name

            value = getattr(self, f.name)

            if isinstance(value, str):
                result[field_name] = value
            elif isinstance(value, NonStandard):
                ns = {}
                for f in dataclasses.fields(NonStandard):
                    v = getattr(value, f.name)
                    if v is not None:
                        ns[f.name] = str(v)

                result = result | ns

            elif value is not None:
                result[field_name] = str(value)

        result['ENTRYTYPE'] = type(self).bibtex_entry_type()
        return result

    def _mandatory_values(self) -> dict[str, Any]:
        """Checks if standard mandatory fields are not None."""
        raise NotImplementedError

    @classmethod
    def bibtex_entry_type(cls):
        """Gets the bibtex entrytype name"""
        for bibtex, internal_type in TYPES.items():
            if internal_type == cls:
                return bibtex
        raise ValueError

    @classmethod
    def generic(cls,
                cite_key: str,
                address: str | None = None,
                annote: str | None = None,
                booktitle: str | None = None,
                author: str | None = None,
                chapter: str | None = None,
                crossref: str | Reference | None = None,
                edition: str | None = None,
                editor: str | None = None,
                howpublished: str | None = None,
                institution: str | None = None,
                journal: str | None = None,
                month: str | None = None,
                note: str | None = None,
                number: str | None = None,
                organization: str | None = None,
                pages: str | int | None = None,
                publisher: str | None = None,
                school: str | None = None,
                series: str | None = None,
                title: str | None = None,
                type: str | None = None,
                volume: str | int | None = None,
                year: str | int | None = None,
                non_standard: NonStandard | None = None,
                scope: dict[str, Reference] | None = None) -> Reference:
        """builds a generic reference, allowing to init each field"""
        instance = cls(cite_key=cite_key,
                       address=address,
                       annote=annote,
                       booktitle=booktitle,
                       author=author,
                       chapter=chapter,
                       crossref=crossref.cite_key if isinstance(crossref, Reference) else crossref,
                       edition=edition,
                       editor=editor,
                       howpublished=howpublished,
                       institution=institution,
                       journal=journal,
                       month=month,
                       note=note,
                       number=number,
                       organization=organization,
                       pages=pages,
                       publisher=publisher,
                       school=school,
                       series=series,
                       title=title,
                       type=type,
                       volume=volume,
                       year=year,
                       non_standard=non_standard,
                       scope=scope)

        if any(v is None for k, v in instance._mandatory_values().items()):
            if all(v is not None for k, v in instance.cross_resolved()._mandatory_values().items()):
                LOG.info('all mandatory values resolved in scope cross references')
            else:
                raise ValueError(f'missing mandatory field for {cls.__name__} {instance.cite_key}')

        # scope management for crossref
        if scope is not None:
            # les chaines vides doivent être ignorées car elles représentent l'unique clef des citations anonymes
            if cite_key != '' and cite_key in scope:
                raise ValueError(f'{cite_key} is already present in bibliograpy scope for {scope[cite_key]}')

            scope[cite_key] = instance

        return instance

    @classmethod
    def from_dict(cls, source: dict[str, Any], scope: dict[str, Reference] | None) -> Reference:
        """Builds a reference from a dict."""
        return cls.generic(
            cite_key=source[Reference.CITE_KEY_FIELD],
            address=source['address'] if 'address' in source else None,
            annote=source['annote'] if 'annote' in source else None,
            author=source['author'] if 'author' in source else None,
            booktitle=source['booktitle'] if 'booktitle' in source else None,
            chapter=source['chapter'] if 'chapter' in source else None,
            crossref=source['crossref'] if 'crossref' in source else None,
            edition=source['edition'] if 'edition' in source else None,
            editor=source['editor'] if 'editor' in source else None,
            howpublished=source['howpublished'] if 'howpublished' in source else None,
            institution=source['institution'] if 'institution' in source else None,
            journal=source['journal'] if 'journal' in source else None,
            month=source['month'] if 'month' in source else None,
            note=source['note'] if 'note' in source else None,
            number=source['number'] if 'number' in source else None,
            organization=source['organization'] if 'organization' in source else None,
            pages=source['pages'] if 'pages' in source else None,
            publisher=source['publisher'] if 'publisher' in source else None,
            school=source['school'] if 'school' in source else None,
            series=source['series'] if 'series' in source else None,
            title=source['title'] if 'title' in source else None,
            type=source['type'] if 'type' in source else None,
            volume=source['volume'] if 'volume' in source else None,
            year=source['year'] if 'year' in source else None,
            non_standard=NonStandard.from_dict(source),
            scope=scope)


@dataclass(frozen=True)
class ReferenceBuilder:
    """A builder of reference decorators."""

    reference_wrapper: Callable[[list[Reference]], str]

    @staticmethod
    def _default_lambda(prefix: str,
                        itemizer: str,
                        reference_formatter,
                        refs: list[Reference]) -> str:

        if len(refs) == 1:
            return f"\n\n{prefix} {reference_formatter(refs[0].cross_resolved())}\n"

        result = f"\n\n{prefix}\n\n"
        for r in refs:
            result += f"{itemizer} {reference_formatter(r.cross_resolved())}\n"
        return result

    @staticmethod
    def default(prefix: str = 'Bibliography:',
                itemize: str = '*',
                reference_formatter = lambda r: (f"{r.title} [{r.cite_key}]" if r.cite_key != _ANONYM_CITE_KEY
                                                                             else r.title)):
        """the default reference decorator"""
        return ReferenceBuilder(
            reference_wrapper=lambda r: ReferenceBuilder._default_lambda(prefix, itemize, reference_formatter, r))

    def __call__(self, *refs):
        """The reference decorator."""

        def internal(obj):
            if obj.__doc__ is None:
                obj.__doc__ = ''
            if len(refs) == 1:
                ref0 = refs[0]
                if isinstance(ref0, Reference):
                    obj.__doc__ += self.reference_wrapper([ref0])
                elif isinstance(ref0, list):
                    obj.__doc__ += self.reference_wrapper(ref0)
            else:
                obj.__doc__ += self.reference_wrapper([*refs])
            return obj

        return internal

cite = ReferenceBuilder.default()

class _InternalReference(Reference):
    """Internal bibliographic usage before defining standard types."""

    def _mandatory_values(self):
        """Checks if standard mandatory fields are not None."""
        return {}

_bibtex_com = cite(_InternalReference.generic(cite_key='bibtex_com',
                                                   title='www.bibtex.com'))

_bibtex_package = cite(
    _InternalReference.generic(cite_key='bibtex_package',
                               title='CTAN Bibtex package documentation',
                               non_standard=NonStandard(
                           url='https://distrib-coffee.ipsl.jussieu.fr/pub/mirrors/ctan/biblio/bibtex/base/btxdoc.pdf')
                               ))

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Article(Reference):
    """any article published in a periodical like a journal article or magazine article

    An article from a journal or magazine.
    Required fields: author, title, journal, year.
    Optional fields: volume, number, pages, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values"""
        return {
            'author': self.author,
            'title': self.title,
            'journal': self.journal,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Book(Reference):
    """a book

    A book with an explicit publisher.
    Required fields: author or editor, title, publisher, year.
    Optional fields: volume or number, series, address, edition, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author or editor': self.author or self.editor,
            'title': self.title,
            'publisher': self.publisher,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Booklet(Reference):
    """like a book but without a designated publisher

    A work that is printed and bound, but without a named publisher or sponsoring institution.
    Required field: title.
    Optional fields: author, howpublished, address, month, year, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {'title': self.title}

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Inbook(Reference):
    """a section or chapter in a book

    A part of a book, which may be a chapter (or section or whatever)and/or a range of pages.
    Required fields: author or editor, title, chapter and/or pages, publisher, year.
    Optional fields: volume or number, series, type, address, edition, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author or editor': self.author or self.editor,
            'title': self.title,
            'chapter or pages': self.chapter or self.pages,
            'publisher': self.publisher,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Incollection(Reference):
    """an article in a collection

    A part of a book having its own title.
    Required fields: author, title, booktitle, publisher, year.
    Optional fields: editor, volume or number, series, type, chapter, pages, address, edition, month, note"""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'booktitle': self.booktitle,
            'publisher': self.publisher,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Inproceedings(Reference):
    """a conference paper (same as the conference entry type)

    An article in a conference proceedings.
    Required fields: author, title, booktitle, year.
    Optional fields: editor, volume or number, series, pages, address, month, organization, publisher, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'booktitle': self.booktitle,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Conference(Inproceedings):
    """The same as INPROCEEDINGS, included for Scribe compatibility."""

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Manual(Reference):
    """a technical manual

    manual Technical documentation.
    Required field: title.
    Optional fields: author, organization, address, edition, month, year, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {'title': self.title}

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Mastersthesis(Reference):
    """a Masters thesis

    mastersthesis A Master’s thesis.
    Required fields: author, title, school, year.
    Optional fields: type, address, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'school': self.school,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Misc(Reference):
    """used if nothing else fits

    misc Use this type when nothing else fits.
    Required fields: none.
    Optional fields: author, title, howpublished, month, year, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {}

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Phdthesis(Reference):
    """a PhD thesis

    phdthesis A PhD thesis.
    Required fields: author, title, school, year.
    Optional fields: type, address, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'school': self.school,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Proceedings(Reference):
    """the whole conference proceedings

    proceedings The proceedings of a conference.
    Required fields: title, year.
    Optional fields: editor, volume or number, series, address, month, organization, publisher, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'title': self.title,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class TechReport(Reference):
    """a technical report, government report or white paper

    techreport A report published by a school or other institution, usually numbered within a series.
    Required fields: author, title, institution, year.
    Optional fields: type, number, address, month, note."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'institution': self.institution,
            'year': self.year
        }

@_bibtex_package
@_bibtex_com
@dataclass(frozen=True, repr=False)
class Unpublished(Reference):
    """a work that has not yet been officially published

    unpublished A document having an author and title, but not formally published.
    Required fields: author, title, note.
    Optional fields: month, year."""

    def _mandatory_values(self):
        """Returns all the mandatory values."""
        return {
            'author': self.author,
            'title': self.title,
            'note': self.note
        }

# Déclarations de références anonymes
# les références anonymes n'ont pas de cite_key
# elles ne sont pas réutilisables et sont simplement déclarées une seule fois

def _anonym(constructor: type[Reference]):
    def internal(
            address: str | None = None,
            annote: str | None = None,
            booktitle: str | None = None,
            author: str | None = None,
            chapter: str | None = None,
            crossref: str | Reference | None = None,
            edition: str | None = None,
            editor: str | None = None,
            howpublished: str | None = None,
            institution: str | None = None,
            journal: str | None = None,
            month: str | None = None,
            note: str | None = None,
            number: str | None = None,
            organization: str | None = None,
            pages: str | int | None = None,
            publisher: str | None = None,
            school: str | None = None,
            series: str | None = None,
            title: str | None = None,
            type: str | None = None,
            volume: str | int | None = None,
            year: str | int | None = None,
            non_standard: NonStandard | None = None,
            ref_supplier = cite):
        return ref_supplier(constructor.generic(
                    cite_key=_ANONYM_CITE_KEY,
                    address=address,
                    annote=annote,
                    booktitle=booktitle,
                    author=author,
                    chapter=chapter,
                    crossref=crossref,
                    edition=edition,
                    editor=editor,
                    howpublished=howpublished,
                    institution=institution,
                    journal=journal,
                    month=month,
                    note=note,
                    number=number,
                    organization=organization,
                    pages=pages,
                    publisher=publisher,
                    school=school,
                    series=series,
                    title=title,
                    type=type,
                    volume=volume,
                    year=year,
                    non_standard=non_standard,
                    scope=None if crossref is None else crossref.scope))
    return internal

article = _anonym(Article)
book = _anonym(Book)
booklet = _anonym(Booklet)
inbook = _anonym(Inbook)
incollection = _anonym(Incollection)
inproceedings = _anonym(Inproceedings)
conference = _anonym(Conference)
manual = _anonym(Manual)
mastersthesis = _anonym(Mastersthesis)
misc = _anonym(Misc)
phdthesis = _anonym(Phdthesis)
proceedings = _anonym(Proceedings)
techreport = _anonym(TechReport)
unpublished = _anonym(Unpublished)

TYPES: dict[str, type[Reference]] = {
    'article': Article,
    'book': Book,
    'booklet': Booklet,
    'inbook': Inbook,
    'incollection': Incollection,
    'inproceedings': Inproceedings,
    'conference': Conference,
    'manual': Manual,
    'mastersthesis': Mastersthesis,
    'misc': Misc,
    'phdthesis': Phdthesis,
    'proceedings': Proceedings,
    'techreport': TechReport,
    'unpublished': Unpublished
}



SHARED_SCOPE: dict[str, Reference] = {}
